/*
   --------------------------------------------------------------------
   BDEIndex.cc
   October 4, 2003, original version 
   November 13, 2003, latest version
   Author: Michael Nuwer

   Allows direct interaction with the BDE through the BDE's API
   to modify index tags for *local tables* (dBASE and Paradox).
   This is not useful for SQL server tables.

   Minor enhancements added by Ken Mayer:
      Changed to use sIDXDescEX at the bottom of the first
      class in this file. 

      Changed prototype of DbiGetIndexDesc to one suggested
      by Jim Sare.

      Added comments on usage.

      Added ability in TagProperties method to store 
      the properties to the object itself.

   Changes:
      November 13, 2003: A minor bug fix made in TagProperties()
                         method. Changed nUnique to bUnique

   Methods:
      AddTag()         -- Add a tag to a local table
      DeleteTag()      -- Delete a tag from a local table
      TagList()        -- List of tag names (returns an array)
      TagNumber()      -- TagNumber (tag number of an index tag
                          -- useful if you know the name, but
                          not the sequence it is stored in the
                          table)
      TagProperties()  -- Returns an associative array of 
                          properties

      For details on usage, see below:

   Usage:
      set procedure to :dUFLP:BDEIndex.cc additive
      oIDX = new BDEIndex()
      // set up your query for your table, if not already
      // done:
      q = new query()
      q.sql = "select * from tablename"
      q.active = true

      // ----------------------------------------------
      // AddTag()
      // Add an index tag to a table -- for this the
      // table should not be opened or in a query:
      oIDX.AddTag( cTablename, oDatabase )
          // where;
          // 'cTableName' is the name of the table
          //    you wish to add the tag to
          // 'oDatabase' is a database object that
          //    already exists for the BDE Alias -- pass
          //    an empty string if not using this

                Special info:
                Before calling this method, define the index
                characteristics by assigning the appropriate
                properties. For example:

                i = new BDEIndex()
                i.indexName  := "Index1"
                i.expression := "Field3"
                i.descending := false
                i.AddTag( cTablename, oDatabase )

                The following is a list for both dBASE and
                Paradox table index properties.
                Required items have a * by them:

                DBASE                     Paradox
                IndexName (char) *        IndexName (char)
                Expression (char) *       Fields (char) *
                ForExpression (char)      DescList(char)
                Primary (logical)         Primary (logical)
                Unique (logical)          Unique (logical)
                Descending (logical)      Descending (logical)
                Distinct (logical)        Maintained (logical)
                                          CaseSensitive (logical)

                dBASE Indexes: 
                  Expression and ForExpression:
                     Regular dBASE expressions, no UDFs, use
                     + to combine fields, etc.

                  A index can not be both distinct and unique. 
                  Use only one of these properties per index.
                                    
                Paradox Indexes:
                  IndexName:
                    May be blank if Fields has a single field
                    in it, otherwise it is required.
                  Fields:
                    No expressions, if multiple fields, use
                    a semicolon as a separator:
                       Field1;Field2;Field3    etc.
                  DescList:
                    Descending for each field in Fields.
                    If this is not used, and descending is true,
                    then all fields in the keyfield list will
                    be descending. If this is used, then
                    you must set true or false for each
                    field in the field list, i.e.,
                       true;false;true;false
                    Where 'true' means the field is descending,
                    and false means it is 'ascending'.

       // This method returns a logical value -- if successful
       // it returns 'true', if it is not successful, it returns
       // false.

      // ----------------------------------------------
      // DeleteTag()
      // Delete an index tag to a table -- for this the
      // table should not be opened or in a query:
      oIDX.DeleteTag( cTablename, oDatabase, cIndexName )
          // where:
          // 'cTableName' is the name of the table
          //    you wish to Delete the tag from
          // 'oDatabase' is a database object that
          //    already exists for the BDE Alias -- pass
          //    an empty string if not using this
          // 'cIndexName' is the tag you wish to delete
      // ----------------------------------------------
      // TagList()
      // Get an array of tagnames:
      aTagNames = oIDX.TagList( q ) 
          // where 'q' is the query object reference
          // returns an empty array if an error occurs

      // ----------------------------------------------
      // TagNumber()
      // Return the index number for any specific 
      // index tag:
      nIndex = oIDX.TagNumber( q, cName )
         // where 'q' is the query object reference
         // and   'cName' is the tag name
         // returns 0 if not found

      // ----------------------------------------------
      // TagProperties()
      // Return information about a specific index tag:
      aTagProps = oIDX.TagProperties( q, nIndex )
         // where 'q' is the query object reference
         // and   'nIndex' is the index NUMBER
          // returns an empty array if an error occurs

      This returns an associative array of properties
      which will vary depending on whether or not the
      index is a .DBF index, or anything else (Paradox,
      SQL server, etc.). This also stores these properties
      to the object itself, using the same names in the list
      below.

      To output these properties to the output pane of
      the command window:
         aTagProps = oIDX.TagProperties( q, nIndex )
         cKey = aTagProps.first // first element:
         for i = 1 to aTagProps.count()
             ? cKey+" = "+aTagProps[ cKey ]
             cKey = aTagProps.nextKey( cKey )
         next

      You can also just check the individual properties:
         oIDX.TagProperties( q, nIndex )
         ? oIDX.IndexName
         ? oIDX.Expression
         // etc.

      To get the same information for another index tag,
      repeat ...

      If you know the name of the property, you can query that
      explicitly. Note that keys in an associative array are
      case-sensitive:

      ? aTagProps[ "Tag Name" ]
      ? aTagProps[ "Expression" ]
      // etc. 

      The properties that are returned are:

      For a DBF Table:           For a PDX or other Table:
      IndexName                  IndexName
      Expression                 Primary
      ForExpression              Unique
      Primary                    Descending
      Unique                     KeyLength
      Distinct                   Maintained
      Descending                 CaseSensitive
      KeyLength                  KeyFieldnn *
                                  (for each field in the expression,
                                   this will note ascending or
                                   descending) 
                                 NumFieldsInKey
                                   // numeric value

                                 * As Properties of this 
                                   object:
                                   Fields (fieldnames in key
                                          separated by ";")
                                   DescList (descending = true,
                                             ascending = false,
                                             separated by ";" for
                                             each field ...)

      // ******** DEPLOYMENT ISSUE:
      // If you deploy this with your own applications,
      // it relies on:

	    vStruct.prg (and vStruct.h)
      
      // You should be sure to include this in your deployment
      // or project ...
   --------------------------------------------------------------------
*/


class BDEIndex 
   // Jim Sare's prototype:
   if type("DbiGetIndexDesc") # "FP"
      extern CUSHORT DbiGetIndexDesc(;  // DBIResult DbiGetIndexDesc
                                      CHANDLE,;  // hDBICur   hCursor
                                      CUSHORT,;  // UINT16    iIndexSeqNo
                                      CPTR);     // pIDXDesc  pidxDesc
        idapi32
   endif

   this.IndexName     = ""       //  Both
   this.Expression    = ""       //  dBASE
   this.ForExpression = ""       //  dBASE
   this.Primary       = false    //  Both   
   this.Unique        = false    //  Both     
   this.Descending    = false    //  Both
   this.Distinct      = false    //  dBASE   
   this.Maintained    = true     //  Paradox     
   this.CaseSensitive = false    //  Paradox
   this.Fields        = ""       //  Paradox    
   this.DescList      = ""       //  Paradox  

   function TagList( oQuery )
      ///////////////////////////////////////
      // Returns an array containing the  
      // tagNames of the table used in 
      // oQuery -- in case of an error,
      // this returns an empty array
      // -- you can check size ...
      ///////////////////////////////////////
      local oIDXDesc, aIndexList, x, n
      aIndexList = new Array()
      if oQuery.rowset.tempTable
         msgbox("The query is using a temp table and has no indexes",;
                "BDEIndex::TagList Error", 16 )
         return aIndexList
      endif

      // use extended IDXDesc class below
      oIDXDesc = new sIDXDescEx()
      x = DbiGetIndexDesc(oQuery.rowset.handle,1,oIDXDesc)
      if x = 0  // At least one index exits
         n = 1
         do
            if oQuery.rowset.tableDriver = "DBASE"
               aIndexList.add( oIDXDesc.szTagName() )
            else
               if n=1 and empty( oIDXDesc.szName() )
                  aIndexList.add( "Primary Key" )
               else
                  aIndexList.add( oIDXDesc.szName() )
               endif
            endif
            n++
            x = DbiGetIndexDesc(oQuery.rowset.handle,n,oIDXDesc)
         until x <> 0
      endif
   return aIndexList

   function TagNumber( oQuery, cTagName )
      //////////////////////////////////////////
      //  Returns the tag number for a
      //  specified tag name. Returns 0 if
      //  the name is not found.
      //////////////////////////////////////////
      local aIndexList, i, nIndex
      nIndex = 0
      if oQuery.rowset.tempTable
         msgbox("The query is using a temp table and has no indexes",;
                "BDEIndex::TagNumber Error", 16 )
         return nIndex
      endif
      aIndexList = this.TagList(oQuery)
      for i = 1 to aIndexList.size
         if upper(aIndexList[i]) == upper(cTagName)
            nIndex = i
            exit
         endif
      next
   return nIndex

   function TagProperties( oQuery, nIndexNumber )
      ////////////////////////////////////
      //  Returns an associative array
      //  containing the properties of 
      //  an index tag. If an error
      //  occurs, it returns an empty
      //  array -- check count() as
      //  this is an associative array
      ////////////////////////////////////
      local oIDXDesc,x,aProperties,cType, i
      aProperties = new assocArray()
      if oQuery.rowset.tempTable
         msgbox("The query is using a temp table and has no indexes",;
                "BDEIndex::TagProperties Error", 16 )
         return aProperties
      endif
      oIDXDesc = new sIDXDescEX()
      x = DbiGetIndexDesc(oQuery.rowset.handle,nIndexNumber,oIDXDesc)
      if x <> 0
         msgbox( "There is no index tag for sequence number " + ;
                  nIndexNumber,;
                 "BDEIndex::TagProperties Error", 16 )
         return aProperties
      endif
      cType = oQuery.rowset.tableDriver
      if cType == 'DBASE' 
         aProperties["IndexName"]        = oIDXDesc.szTagName()
         this.IndexName                  = oIDXDesc.szTagName()
         aProperties["Primary"]          = oIDXDesc.bPrimary()
         this.Primary                    = oIDXDesc.bPrimary()
         aProperties["Unique"]           = oIDXDesc.bUnique()
         this.Unique                     = oIDXDesc.bUnique()
         aProperties["Distinct"]         = oIDXDesc.bDistinct()
         this.Distinct                   = oIDXDesc.bDistinct()
         aProperties["Descending"]       = oIDXDesc.bDescending()
         this.Descending                 = oIDXDesc.bDescending()
         aProperties["KeyLength"]        = oIDXDesc.iKeyLen()
         this.KeyLength                  = oIDXDesc.iKeyLen()
         aProperties["ForExpression"]    = oIDXDesc.szKeyCond()
         this.ForExpression              = oIDXDesc.szKeyCond()
         aProperties["Expression"]       = oIDXDesc.szKeyExp()
         this.Expression                 = oIDXDesc.szKeyExp()
      else  // Paradox, InterBase, ODBC 
         aProperties["IndexName"]         = oIDXDesc.szName( )
         this.IndexName                   = oIDXDesc.szName( )
         aProperties["Primary"]           = oIDXDesc.bPrimary()
         this.Primary                     = oIDXDesc.bPrimary()
         aProperties["Unique"]            = oIDXDesc.bUnique()
         this.Unique                      = oIDXDesc.bUnique()
         aProperties["Descending"]        = oIDXDesc.bDescending()
         this.Descending                  = oIDXDesc.bDescending()
         aProperties["KeyLength"]         = oIDXDesc.iKeyLen()
         this.KeyLen                      = oIDXDesc.iKeyLen()
         aProperties["Maintained"]        = oIDXDesc.bMaintained()
         this.Maintained                  = oIDXDesc.bMaintained()
         aProperties["CaseSensitive"]     = Not oIDXDesc.bCaseInsensitive() 
         this.CaseSensitive               = Not oIDXDesc.bCaseInsensitive() 
         aProperties["NumFieldsInKey"]    = oIDXDesc.iFldsInKey()
         this.NumFieldsInKey              = oIDXDesc.iFldsInKey()
         this.Fields                      = ""
         this.DescList                    = ""
         for i = 1 To oIDXDesc.iFldsInKey() 
            cKeyName = "KeyField"+str(i,2,0,"0")
            aProperties[cKeyName] = ;
                 oQuery.rowset.fields[oIDXDesc.aiKeyFld(i)].fieldName
            this.fields += oQuery.rowset.fields[oIDXDesc.aiKeyFld(i)].fieldName
            if i < oIDXDesc.iFldsInKey()
               this.fields += ";"
            endif
            //  If the field is in descending order
            //  attach a note to the fieldname.
            //  Paradox tables only. 
            if cType == 'PARADOX' 
               if oIDXDesc.abDescending(i)  
                  aProperties[cKeyName] += " (Descending)"
                  this.DescList += "true"
               else
                  aProperties[cKeyName] += " (Ascending)"
                  this.DescList += "true"
               endif
               if i < oIDXDesc.iFldsInKey()
                  this.DescList += ";"
               endif
            endif
         endFor
      endif
   return ( aProperties )

   function AddTag(cTablename, oDatabase )
      //////////////////////////
      // Add an index tag to a table
      //////////////////////////
      local cType,oIDXDesc,oErrorCheck,oQuery,aTagList
      local bDoesPrimaryExist,nRet,n,i
      local nFldCount,cFieldList,cField,cList,cValue

      if empty( oDatabase )
         nDBHandle = _app.databases[1].handle
      else
         nDBHandle = oDatabase.handle
      endif

      // determine the table type
      if lower(substr(cTablename,at(".",cTablename))) == ".dbf"
         cType = "DBASE"
      elseif lower(substr(cTablename,at(".",cTablename))) == ".db"
         cType = "PARADOX"
      else
         msgbox("Error: Unknown table type. A file extension is required.",;
                "BDEIndex::AddTag Error", 16 )
         return false
      endif 

      //  Proto-type the function
      if type("DbiAddIndex") # "FP"
         extern cint DbiAddIndex(;
             cHandle, ;
             cHandle, ; 
             cString, ;
             cString, ;
             CPTR, ;
             cString) idapi32
      endif

      oIDXDesc = new sIDXDescEx()
      oErrorCheck = new sIDXDescEx()
      oQuery = new query()
      if not empty(oDatabase)
         oQuery.database = oDatabase
      endif
      oQuery.sql = [select * from "] + cTablename +["]
      oQuery.active = true

      //  Check to see whether a Primary Index exists
      //  And create a tag list to check new name against.
      aTagList = new Array()
      bDoesPrimaryExist = false
      nRet = DbiGetIndexDesc(oQuery.rowset.handle,1,oErrorCheck)
      if nRet = 0  // At least one index exits
         n = 1
         do
            if oQuery.rowset.tableDriver = "DBASE"
               aTagList.add( lower(oErrorCheck.szTagName()) )
            else
               aTagList.add( lower(oErrorCheck.szName()) )
            endif

            if oErrorCheck.bPrimary()
               bDoesPrimaryExist = true
            endif

            n++
            nRet = DbiGetIndexDesc(oQuery.rowset.handle,n,oErrorCheck)
         until nRet <> 0
      endif

      n=null 
      if cType = 'DBASE'

         local cTagName,cExpression,cForCondition,bPrimary
         local bUnique,bDescending,bDistinct
         //  Set property defaults
         cTagName      = ""
         cExpression   = ""
         cForCondition = ""
         bPrimary      = false
         bUnique       = false
         bDescending   = false
         bDistinct     = false

         if this.Unique and this.Distinct
            msgbox("Error: Can not create an index that is both distinct and unique.",;
                      "BDEIndex::AddTag Error", 16 )
            return false
         endif

         cExpression   = this.Expression
         cTagName      = this.indexName
         cForCondition = this.ForExpression
         bPrimary      = this.Primary
         bDescending   = this.Descending
         if this.Unique
            bUnique    = true
         endif
         if this.Distinct
            bDistinct  = true
         endif              

         if bPrimary
            if bDoesPrimaryExist
               msgbox("Error: A primary key already exits.",;
                      "BDEIndex::AddTag Error", 16 )
               return false
            else
               bUnique = true
            endif
         endif

         //  Determine whether we are dealing with a 
         //  compound index or a single field index
         if "+" $ cExpression or "(" $ cExpression
            //  if a compound expression  
         
            oIDXDesc.bExpIdx(true)
            oIDXDesc.szKeyExp( cExpression )  
            oIDXDesc.iFldsInKey(0)
            oIDXDesc.aiKeyFld(1,0)

         else
            //  if a single field expression
            oIDXDesc.bExpIdx(false)
            oIDXDesc.szKeyExp( "" )  
            oIDXDesc.iFldsInKey(1)  

            //  We must now find the fieldNumber for
            //  for the index field.
            for i=1 to oQuery.rowset.fields.size
               if lower(oQuery.rowset.fields[i].fieldname) == lower(cExpression)
                  oIDXDesc.aiKeyFld(1,i)
                  if empty(cTagName)
                     // if user did not pass a tagname
                     // use the field name
                     cTagName = oQuery.rowset.fields[i].fieldName
                  endif
                  exit
                endif
            next
         endif

         ///  Determine the Tag Name
         if empty(cTagName)
            msgbox("A TagName for this index must be provided.",;
                   "BDEIndex::AddTag Error", 16 )
         else

            if aTagList.scan(lower( cTagName )) > 0
               msgbox("Error: " +cTagName+ " is a duplicate tag name.",;
                      "BDEIndex::AddTag Error", 16 )
               return false
            else
               oIDXDesc.szTagName(cTagName)
            endif	
         endif

         oIDXDesc.bPrimary(bPrimary)

         //  Set the next two propeties only if they
         //  are true. Never set one true and the other
         //  false. Because the false one will nulify
         //  the true one. (A strange issue with the BDE.)
         if bUnique
            oIDXDesc.bUnique(true)
         endif
         if bDistinct
            oIDXDesc.bDistinct(true)
         endif

         oIDXDesc.bDescending(bDescending)

         if not empty(cForCondition)
            oIDXDesc.bSubset(true)
            oIDXDesc.szKeyCond( cForCondition )  
         else   
            oIDXDesc.bSubset(False)
            oIDXDesc.szKeyCond( '' )  
         endif

         oIDXDesc.iIndexId(0)
         oIDXDesc.bMaintained(True)
         oIDXDesc.bCaseInsensitive(False)
         oIDXDesc.iBlockSize( 0 )

      elseif cType = "PARADOX"
         //  get properties from array

         local bUnique,bMaintained,bCaseSensitive
         local bPrimary,cKeyFld,bDescending,cDescList

         bUnique        = false
         bMaintained    = true 
         bCaseSensitive = false
         bPrimary       = false
         cKeyFld        = ""   // can be empty for a Primary Key
         bDescending    = false
         cDescList      = ""
         
         bCaseSensitive = this.CaseSensitive
         bMaintained    = this.Maintained
         bUnique        = this.Unique
         bPrimary       = this.Primary
         cKeyFld        = this.Fields
         //  We should use the aiDescending array but
         //  dBASE uses this property, so we will too.
         bDescending    = this.Descending
         cDescList      = this.DescList

         if bPrimary
            if bDoesPrimaryExist
               msgbox("Error: A primary key already exits.",;
                      "BDEIndex::AddTag Error", 16 )
               return false
            endif
            //  Make a primary key
            oIDXDesc.iIndexId(0)
            oIDXDesc.bPrimary(True)
            oIDXDesc.bUnique(True)
            oIDXDesc.bDescending(False)
            oIDXDesc.bMaintained(True)
            oIDXDesc.bSubset(False)
            oIDXDesc.bExpIdx(False)

            oIDXDesc.bCaseInsensitive(False)

            //  handle Primary Key with 
            //  multiple fields
            if empty(cKeyFld)  // no field passed
                               // use field one. 
               oIDXDesc.iFldsInKey(1)
               oIDXDesc.aiKeyFld(1,1)

            else

               //  Field1;Field2;Field3
               if rat(";",cKeyFld) # 1 
                  // There is no trailing semicolon 
                  cFieldList = cKeyFld + ";"
               else
                  cFieldList = cKeyFld
               endif
               nFldCount = 0
               do while at(";",cFieldList) > 0
                  //  get the next field name
                  cField = left(cFieldList,at(";",cFieldList)-1)
                  nFldCount++
                  // find the field number
                  for i=1 to oQuery.rowset.fields.size
                     if lower( oQuery.rowset.fields[i].fieldname) == lower(cField) 

                         oIDXDesc.aiKeyFld(nFldCount,i)
  
                         // we found this one so: 
                         exit

                      endif  // field found
                  next // field 

                  // drop current field from field list
                  cFieldList = substr(cFieldList,at(";",cFieldList)+1)
               enddo // next field list item

               oIDXDesc.iFldsInKey(nFldCount)
            endif
         else
            // Secondary index
            if not bDoesPrimaryExist
               msgbox("Error: A primary key must be added before adding a secondary index.",;
                      "BDEIndex::AddTag Error", 16 )
               return false
            endif
            if empty(cKeyFld) 
               msgBox("Error: At least one field must be used for the key.",;
                      "BDEIndex::AddTag Error", 16 ) 
               return false  
            endif

            oIDXDesc.bPrimary(False)
            oIDXDesc.bUnique(bUnique)
            oIDXDesc.bMaintained(bMaintained)
            oIDXDesc.bCaseInsensitive(not bCaseSensitive)

            //  Deal with the field list
            //  Construct the aiKeyFld array
            //  aiKeyFld( nElement, nField )

            //  Field1;Field2;Field3
            if rat(";",cKeyFld) # 1 
               // There is no trailing semicolon 
               cFieldList = cKeyFld + ";"
            else
               cFieldList = cKeyFld
            endif
            nFldCount = 0
            do while at(";",cFieldList) > 0
               //  get the next field name
               cField = left(cFieldList,at(";",cFieldList)-1)
               nFldCount++
               // find the field number
               for i=1 to oQuery.rowset.fields.size
                  if lower( oQuery.rowset.fields[i].fieldname) == lower(cField) 

                      oIDXDesc.aiKeyFld(nFldCount,i)

                      exit

                   endif  // field found
               next // field 

               // drop current field from field list
               cFieldList = substr(cFieldList,at(";",cFieldList)+1)
            enddo // next field list item

            oIDXDesc.iFldsInKey(nFldCount)


            //  Deal with the index name
            if not empty(this.indexName ) 
               cName = this.indexName
            else
               msgBox("Error: No index name was given.",;
                      "BDEIndex::AddTag Error", 16 )
               return false
            endif

            // check for duplicate name
            if aTagList.scan(lower( cName )) > 0
               msgbox("Error: " +cName+ " is a duplicate index name.",;
                      "BDEIndex::AddTag Error", 16 )
               return false
            else
               oIDXDesc.szName(cName)
            endif
            
            if not empty(cDescList)
               //  the Paradox method. dBASE has no UI way
               //  to set this list.    

               if  rat(";",cDescList) # 1 
                  // There is no trailing semicolon 
                  cList = cDescList + ";"
               else
                  cList = cDescList
               endif
           
               // true;false;true
               for i=1 to nFldCount
                  cValue = left(cList,at(";",cList)-1)
                  if cValue = 'true'
                     oIDXDesc.abDescending(i,true)
                  else
                     oIDXDesc.abDescending(i,false)
                  endif
                  cList = substr(cList,at(";",cList)+1)
               next

               if "true" $ lower(cDescList)
                  oIDXDesc.bDescending(true)
               else
                  oIDXDesc.bDescending(false)
               endif
            else
               //  No DescList and Descending is true
               //  Set all Key fields to descending.
               if bDescending 
                  for i=1 to nFldCount
                     oIDXDesc.abDescending(i,true)
                  next
                  oIDXDesc.bDescending(true)
               else
                  oIDXDesc.bDescending(false)
               endif
            endif  //  DescList
         endif //  Secondary index
         oIDXDesc.iIndexId(0)
         oIDXDesc.bSubset(False)
         oIDXDesc.bExpIdx(false)   
      else  
         msgbox("Error: There was an unknown error.",;
                "BDEIndex::AddTag Error", 16 )
         return false
      endif

      //  Close the table before adding the index
      //  Exclusive use is required.
      oQuery.active = false
      oQuery = null

      ///  Make the index -- hurray
      ErrorCode = DbiAddIndex(nDBHandle, ;
                NULL, ;
                cTablename, ;
                NULL, ;
                oIDXDesc.string, ;
                NULL )

      //  check for an error
      if ErrorCode > 0
         if type("DbiGetErrorString") # "FP"
            extern cint DbiGetErrorString(;
                CLONG, cString) idapi32
         endif
      
         ErrorString = SPACE(128)
         DbiGetErrorString(ErrorCode, ErrorString)
         msgBox("Error: " + ErrorString,;
                "BDEIndex::AddTag Error", 16 )
         return false
      endif

      return true

   function DeleteTag(cTablename, oDatabase, cIndexName )
      ///////////////////////////
      // Delete index tag from a
      // local table
      ///////////////////////////
      /*
      DBIResult DBIFN DbiDeleteIndex (
         hDb,               Type: hDBIDb 	
         hCursor,           Type: hDBICur
         pszTableName,      Type: pCHAR 	
         [pszDriverType],   Type: pCHAR 	
         pszIndexName,      Type: pCHAR 	
         pszIndexTagName,   Type: pCHAR
         iIndexId);         Type: UINT16 
      */
      if type("DbiDeleteIndex") # "FP"
         extern cint DbiDeleteIndex(;
             cHandle, ;
             cHandle, ; 
             cString, ;
             cString, ;
             cString, ;
             cString, ;
             CPTR ) idapi32
      endif

      if empty( oDatabase )
         nDBHandle = _app.databases[1].handle
      else
         nDBHandle = oDatabase.handle
      endif

      // determine the table type
      if lower(substr(cTablename,at(".",cTablename))) == ".dbf"
         cType = "DBASE"
      elseif lower(substr(cTablename,at(".",cTablename))) == ".db"
         cType = "PARADOX"
      else
         msgbox("Error: Unknown table type. A file extension is required.",;
                "BDEIndex::DeleteTag Error", 16 )
         return false
      endif 

      if cType = "DBASE"
         cFileName = left(cTableName,at(".",cTablename)-1) + ".mdx"
         ErrorCode = DbiDeleteIndex(nDBHandle, ;
                NULL, ;
                cTablename, ;
                NULL, ;
                cFileName, ;
                cIndexName, ;
                NULL )
      elseif cType = "PARADOX"
         if not empty(cIndexName)
            ErrorCode = DbiDeleteIndex(nDBHandle, ;
                NULL, ;
                cTablename, ;
                NULL, ;
                cIndexName, ;
                NULL, ;
                NULL )
         else
            // Primary key
            ErrorCode = DbiDeleteIndex(nDBHandle, ;
                NULL, ;
                cTablename, ;
                NULL, ;
                NULL, ;
                NULL, ;
                 0 )
         endif

      else
         msgbox("Error: There was an unknown error.",;
                "BDEIndex::DeleteTag Error", 16 )
      endif

      //  check for an error
      if ErrorCode > 0
         if type("DbiGetErrorString") # "FP"
            extern cint DbiGetErrorString(;
                CLONG, cString) idapi32
         endif
      
         ErrorString = SPACE(128)
         DbiGetErrorString(ErrorCode, ErrorString)
         msgBox("Error: " + ErrorString,;
                "BDEIndex::DeleteTag Error", 16 )
         return false
      endif

   return true

endclass

CLASS sIDXDescEx of VStruct( 893 ) from :dUFLP:vStruct.prg
// Adapted from Jim Sare's stBDE32.cc for use 
// with VStruct

   function szName(c)

      if PCount() > 0
         this.SetStringz(0, c, 261 )
      endif
      return this.GetStringz(0, 261 )

   function iIndexID(n)

      if PCount() > 0
         this.SetShort(261, n)
      endif
      return this.GetShort(261, false)

   function szTagName(c)

      if PCount() > 0
         this.SetStringz(263, c, 32 )
      endif
      return this.GetStringz(263, 32 )

   function szFormat(c)

      if PCount() > 0
         this.SetStringz(295, c, 32 )
      endif
      return this.GetStringz(295, 32 )

   function bPrimary(l)

      if PCount() > 0
         this.SetBoolean16(327, l )
      endif
      return this.GetBoolean16(327)

   function bUnique(l)

      if PCount() > 0
         this.SetShort(329, IIf(l, 1, 0))
      endif
      return this.GetShort(329,false) # 0 and this.GetShort(329,false) # 33

   function bDistinct(l)

      if PCount() > 0
         this.SetShort(329, Iif(l, 33, 0))
      endif
      return this.GetShort(329,false) # 0 and this.GetShort(329,false) # 1

   function bDescending(l)

      if PCount() > 0
         this.SetBoolean16(331, l )
      endif
      return this.GetBoolean16(331) 

   function bMaintained(l)

      if PCount() > 0
         this.SetBoolean16(333, l)
      endif
      return this.GetBoolean16(333)

   function bSubset(l)

      if PCount() > 0
         this.SetBoolean16(335, l)
      endif
      return this.GetBoolean16(335) 

   function bExpIdx(l)

      if PCount() > 0
         this.SetBoolean16(337, l)
      endif
      return this.GetBoolean16(337) 

   function iCost(n)

      if PCount() > 0
         this.SetShort(339, n)
      endif
      return this.GetShort(339, false)

   function iFldsInKey(n)

      if PCount() > 0
         this.SetShort(341, n)
      endif
      return this.GetShort(341, false)

   function iKeyLen(n)

      if PCount() > 0
         this.SetShort(343, n)
      endif
      return this.GetShort(343, false)

   function bOutOfDate(l)

      if PCount() > 0
         this.SetBoolean16(345, l)
      endif
      return this.GetBoolean16(345)

   function iKeyExpType(n)

      if PCount() > 0
         this.SetShort(347, n)
      endif
      return this.GetShort(347, false)

   function aiKeyFld(i, n)

      if (PCount() > 0) And (i > 0) And (i < 17)
         if PCount() > 1
            this.SetShort(349 + ((i - 1) * 2), n)
         endif
         return this.GetShort(349 + ((i - 1) * 2), false)
      endif
      return -1

   function szKeyExp(c)

      if PCount() > 0
         this.SetStringz(381, c, 221)
      endif
      return this.GetStringz(381, 221 )

   function szKeyCond(c)

      if PCount() > 0
         this.SetStringz(602, c, 221)
      endif
      return this.GetStringz(602, 221)

   function bCaseInsensitive(l)

      if PCount() > 0
         this.SetBoolean16(823, l)
      endif
      return this.GetBoolean16(823) 

   function iBlockSize(n)

      if PCount() > 0
         this.SetShort(825, n)
      endif
      return this.GetShort(825, false)

   function iRestrNum(n)

      if PCount() > 0
         this.SetShort(827, n)
      endif
      return this.GetShort(827, false)

   function abDescending(i, l)

      if (PCount() > 0) And (i > 0) And (i < 17)
         if PCount() > 1
            this.SetBoolean16(829 + ((i - 1) * 2), l)
         endif
         return this.GetBoolean16(829 + ((i - 1) * 2)) 
      endif
      return False

   function iUnused(i, n)

      if (PCount() > 0) And (i > 0) And (i < 17)
         if PCount() > 1
            this.SetShort(861 + ((i - 1) * 2), n)
         endif
         return this.GetShort(861 + ((i - 1) * 2), false)
      endif
      return -1

   procedure LogoutData
      local i

      ? "szName:           " + this.szName()
      ? "iIndexID:         " + this.iIndexID()
      ? "szTagName:        " + this.szTagName()
      ? "szFormat:         " + this.szFormat()
      ? "bPrimary:         " + this.bPrimary()
      ? "bUnique:          " + this.bUnique()
      ? "bDistinct:        " + this.bDistinct()
      ? "bDescending:      " + this.bDescending()
      ? "bMaintained:      " + this.bMaintained()
      ? "bSubset:          " + this.bSubset()
      ? "bExpIdx:          " + this.bExpIdx()
      ? "iCost:            " + this.iCost()
      ? "iFldsInKey:       " + this.iFldsInKey()
      ? "iKeyLen:          " + this.iKeyLen()
      ? "bOutOfDate:       " + this.bOutOfDate()
      ? "iKeyExpType:      " + this.iKeyExpType()
      ? "aiKeyFld       " + "1: " + this.aiKeyFld(1)
      for i = 2 To 16
         ? "              " + IIf(i < 10, " ", "") + i + ": " + this.aiKeyFld(i)
      endfor
      ? "szKeyExp:         " + this.szKeyExp()
      ? "szKeyCond:        " + this.szKeyCond()
      ? "bCaseInsensitive: " + this.bCaseInsensitive()
      ? "iBlockSize:       " + this.iBlockSize()
      ? "iRestrNum:        " + this.iRestrNum()
      ? "abDescending   " + "1: " + this.abDescending(1)
      for i = 2 To 16
         ? "              " + IIf(i < 10, " ", "") + i + ": " + this.abDescending(i)
      endfor
      ? "iUnused        " + "1: " + this.iUnused(1)
      for i = 2 To 16
         ? "              " + IIf(i < 10, " ", "") + i + ": " + this.iUnused(i)
      endfor

ENDCLASS  

