/*

	Filename:   FinancialEF.cc
   Version:    1.01 
   Date:       2002 12 27

   For:        dBASE PLUS / dB2K / Visual dBase 7.** 
               Windows 95/98/ME/XP/NT/2000

   Author:     Ronnie MacGregor
   Contact:    FEF_FeedBack@high-lugtonridge.co.uk
	
   
   Description:

      A custom numeric entryfield intended for financial use with:

         Right hand numeric entry and display
          
         No mouse selection of it's contents allowed

         SelectAll = true type behaviour

         No decimal point entry required

         Digits entered or deleted move accross the decimal point as required

         Data format is validated and constricted by the EntryField picture
         property if there is one, otherwise by the properties of the
         datalinked field.

         Automatic formatted zero insertion if the first entered character is
         a decimal point or a minus sign.

      Backspace key deletes most recently entered character.

      Ordinary delete key deletes the whole entry.

      On opening the control creates it's own Text control which covers the
      EntryField. This prevents user mouse access to the EntryField itself,
      and is used to display the control contents.


   Revisions:

      2003/01/01 Version 1.01

      Macro instantiation of CoverText object dropped in favour of creating the
      CoverText as a property of the EntryField while still using the form as
      the container. Slightly simpler, and slightly faster to open.

      The Control will now fire it's onChange event code (if it exists).

      The control will now try to add itself, as a reference, to a custom array
      existing as a property of it's own datalinked rowset (if it exists). This
      array is then available to the developer so that multiple instances of
      this control can all easily be refreshed by using code similar to that
      given as an example under the Methods heading.


   Methods:

.     function Refresh()

      This method must be explicitly called after rowset navigation, saving,
      abandoning, deleteing, or appending, to refresh the control.

      This custom control will try to add itself to it's own datalinked rowset's
      custom array of FinancialEF references, so that multiple instances of this
      control can all be refreshed by using code similar to the following:

.        function rowset_event()
            for nCount = 1 to this.aFinancialEFs.size
               this.aFinancialEFs[nCount].Refresh()
            endfor
            return


   Dependancies:

      None


   Acknowledgements:

      This control was derived from NumericEF.cc which was a solution
      created for Peter Balfour in the dBASE newsgroups, with some code
      taken from his original idea.

      Some elements of the cover control instantiation method derived from
      a sample by Marko Mihorko.
         
*/



class FinancialEF(parentObj) of EntryField(parentObj) custom


   // Ensure control drops onto form correctly
   form.metric := 6 // Pixels

   // Standard and default properties
   with (this)
      onOpen = CLASS::onOpen
      canClose = CLASS::canClose
      onGotFocus = CLASS::onGotFocus
      onLostFocus = CLASS::onLostFocus
      key = CLASS::Key
      height = 22
      width = 100
      value = ""
   endwith
   
   // Custom properties
   this.Version = 1.01     // Version number
   this.ExistingChars = "" // The existing string of accepted characters
   this.HasChanged = false // Whether the control contents have changed
   this.nLength = 0        // Length of picture property or field
   this.nInt = 0           // Integer portion of picture property or field
   this.nDec = 0           // Decimal portion of picture property or field


   // Prevent unwanted outside access to internals
   protect UpdateControl(), AddChar(), DelChar(), FireOnChange()
   protect this.Version, this.ExistingChars, this.HasChanged
   protect this.nLength, this.nInt, this.nDec 

   ////////////////////////////////////////
   

   function onOpen()

      // Create Cover control
      this.CoverText = new TEXT(this.parent)
      with (this.CoverText)
         onLeftMouseDown = {; this.Owner.setFocus()}
         top = this.top
         left = this.left 
         width = this.width
         height = this.height
         border = true
         colorNormal = "WindowText/Window"
         alignHorizontal = 2 // Right
         text = ""
      endwith

      // Give cover a reference to this control
      this.CoverText.Owner = this

      // Deal with picture property
      if len(this.picture) > 0
         this.nLength := len(this.picture)
         this.nDec := len(this.picture) - at(".", this.picture)
         this.nInt := len(this.picture) - this.nDec - 1
      else
         try
            this.nLength := this.datalink.length
            this.nDec := this.datalink.DecimalLength
            this.nInt := this.datalink.length - this.nDec - 1
         catch (exception e)
            // Datalink doesn't exist
            this.nLength := 20
            this.nDec := 2
            this.nInt := 17
         endtry
      endif

      // Initialise display
      this.Refresh()

      // Add a reference this control to the 
      // custom rowset array of FinancialEF controls
      try
         r = this.datalink.parent.parent
      catch (exception e)
         return  // Datalink doesn't exist 
      endtry
      try
         a = r.aFinancialEFs
      catch (exception e)
         // Array doesn't exist so create it
         r.aFinancialEFs = new array()
      endtry
      r.aFinancialEFs.add(this)

      return


   /////////


   function onGotFocus()
      this.start := true
      this.HasChanged := false
      this.CoverText.colornormal := "HighLightText/HighLight"
      return


   /////////


   function onLostFocus()
      this.CoverText.colornormal := "WindowText/Window"
      this.FireOnChange()
      return


   /////////
   

   function canClose()

      // Remove circular reference
      this.CoverText.Owner := null

      // Remove the reference to this control in the 
      // rowset's custom array of FinancialEF controls
      try
         r = this.datalink.parent.parent
         a = r.aFinancialEFs
         for nCount = 1 to  a.size
            if a[nCount].name == this.name
               a.delete(nCount)
            endif
         endfor // next
      catch (exception e)
         return true // entry doesn't exist 
      endtry

      return true


   /////////


   function Key(nChar, nRepCnt, nFlags)

      // Deal with return or tab key
      if nChar == 9
         return false
      endif

      // Deal with Ordinary Delete key
      if nChar == 127
         c = this.ExistingChars
         do     
            c = this.DelChar(c)
         until val(c) == 0
         this.UpdateControl(c)
         return false
      endif

      // Deal with BackSpace
      if nChar == 8
         c = this.ExistingChars
         if len(c) > 0
            c = this.DelChar(c)
            this.UpdateControl(c)
         else
            ? chr(7)
         endif
         return false
      endif

      // Prevent unwanted decimal point
      if nChar == 46 
         ? chr(7)
         return false
      endif

      // Prevent Alpha Entry
      if nChar > 47 and nChar < 58 or nChar == 45 or nChar == 127
         // Allowable key - just continue
      else
         ? chr(7)
         return false
      endif

      // force over-write 
      if this.start == true 
         this.ExistingChars := ""
         this.start = false
      endif

      // Get existing entry 
      c = this.ExistingChars

      // Prevent unwanted minus signs
      if nChar == 45 and len(c) > 0
         ? chr(7)
         return false
      endif

      // Limit key entry
      if this.nDec > 0
         if len(c) >= this.nLength
            ? chr(7)
            return false
         endif
      else
         if len(c) >= this.nInt
            ? chr(7)
            return false
         endif
      endif

      // Add entered Key
      c = this.AddChar(c, nChar)

      // Update Control
      this.UpdateControl(c)

      return false


   /////////

   
   function UpdateControl(c)

      // Update Control Values
      this.HasChanged := true
      this.ExistingChars := c
      this.value := val(c)
      try
         this.datalink.parent.parent.modified := true
      catch (exception e)
         // No datalink
      endtry

      // Update Display
      this.CoverText.text := c

      return


   /////////


   function AddChar(c, nChar)

      do case
       
         case nChar == 45 and len(c) == 0 // "-"

            c = "-0." + replicate("0",this.nDec)
            return c

         case nChar == 46 and len(c) == 0 // "."

            c = "0." + replicate("0",this.nDec)
            return c


         case this.nDec > 0

            sOrig = c

            if not "." $ sOrig

               c = "0." + replicate("0",(this.nDec - len(sOrig) - 1)) + chr(nChar)
               return c

            else

               nPointPos = at(".", sOrig)

               c = substr(sOrig, 1, nPointPos - 1) // Existing integer portion

               if substr(c, 1, 1) == "-"
                  c = substr(c, 2)
                  do while substr(c, 1, 1) == "0"
                     c = substr(c, 2)
                  enddo
                  c = "-" + c
               else
                  do while substr(c, 1, 1) == "0"
                     c = substr(c, 2)
                  enddo
               endif

               c = c + substr(sOrig, nPointPos + 1, 1) // First existing decimal
               c = c + "."
               c = c + substr(sOrig,nPointPos+2,this.nDec - 1) + chr(nChar)

               return c

            endif

         otherwise

            c = c + chr(nChar)

      endcase

      return c


   /////////


   function DelChar(c)

      if "." $ c

         sOrig = c
         nPointPos = at(".", sOrig)

         c = substr(sOrig,1,nPointPos-2)

         if c == "-"
            c = "-0"
         endif
         if len(c) < 1 
            c = c + "0."
         else
            c = c + "."
         endif

         sDecPart = substr(sOrig,nPointPos-1,1) + substr(sOrig,nPointPos+1,1)

         do while len(sDecPart) < this.nDec

            sDecPart = "0" + sDecPart

         enddo

         c = c + sDecPart   

      else

         c = left(c,(len(c))-1)

      endif

      if val(c) == 0
         if this.nDec > 0
            c = "0." + replicate("0",this.nDec)
         else
            c = "0"
         endif
      endif
            
      return c


   /////////


   function FireOnChange()
      if this.HasChanged == true
         try
            this.onChange()
         catch (exception e)
            // No onChange() code
         endtry
         this.HasChanged := false
      endif
      return


   /////////


   function Refresh()
      this.FireOnChange()
      try
         c = ltrim(str(this.datalink.value, this.nLength, this.nDec))
         if c = null
            throw e
         endif
         if len(this.picture) > 0
            if val(c) <> this.datalink.value
               c = replicate("*",this.nInt)
               if this.nDec > 0
                  c = c + "." + replicate("*",this.nDec)
               endif
               this.CoverText.text := c
            endif
         endif
         this.CoverText.text := c
      catch (exception e)
         // Datalink doesn't exist or null table entry
         this.CoverText.text := ""
      endtry
      this.Start = true
      return


   /////////

 
endclass

//eof

