/*
   This program contains code that you may wish to attach to a form,
   a custom form, whatever. It is not independant code, and should
   not be run outside of a form. These are events designed to
   be part of a form, and called from said form. It is suggested
   that if you wish to use these functions, place them before
   the ENDCLASS statement for your form (or even subform), and
   call them as custom methods of the form (i.e., class::Repaint(), 
   and so on ...).

   BringControlToTop() -- Romain Strieff
   EnableControls()    -- Jim Sare
   Repaint()           -- Jim Sare
   Restore()           -- Romain Strieff
   IsKeyPressed()      -- Jim Sare
   Freeze() and
   UnFreeze()          -- Unknown
   ReleaseQueries      -- Jan Hoelterling

   Basically, copy the required function into a form's constructor
   code (before the ENDCLASS statement) ...
*/

msgbox( "This program is not meant to be executed, "+;
        "but contains useful API calls and other functions.","Don't run this!", 16 )
return

/*
   Posted by Romain Strieff in the dBASE, Inc. newsgroups,
   June 23, 1999

   Form.BringControlToTop(form.image1.hwnd)

*/
procedure BringControlToTop(Controlhwnd)
   #if __vdb__ >= 7
      #define USERDLL  USER32
   #else
      #define USERDLL  USER
   #endif
   if type("BringWindowToTop") # "FP"
      extern CINT BringWindowToTop (CINT)  USERDLL
  endif 
return BringWindowToTop(Controlhwnd)
// eof: BringControlToTop()

/*
     Enable Controls -- this will loop through all controls
     on a form and set their enabled property to true or false
     depending on the parameters you pass it.

     Author: Jim Sare
     Posted on the VdBASE Newsgroups, 7/9/1998

     Usage Notes:

        This particular method enables/disables all controls in the 
        form and any containers on the form. It can be adapted to
        other purposes.

        To use it:

        Copy the method into the form class (before the ENDCLASS statement)

        FORM.EnableControls( false ) // disable all controls
        FORM.EnableControls( true )  // enable all controls
        FORM.EnableControls( false, FORM.Notebook1 )
                                     // disable all controls ONLY
                                     // on Notebook1
        FORM.EnableControls( true, FORM.Notebook1 )
                                     // enable all controls ONLY
                                     // on Notebook1
        These last two commands can be used for a container as well.

        The function is recursive so if a container contains
        controls and another container or notebook, the controls
        contained in those containers and/or notebooks will also
        be affected.

*/
PROCEDURE EnableControls
   parameters lEnable, o
   private oRef
   local oParent

   oParent = iif( type( "o" ) == "O", o, FORM )
   oRef = oParent.first
   do
      if type( "oRef.first" ) == "O" 
         form.EnableControls( lEnable, oRef )
      endif
      if type( "oRef.Enabled" ) # "U"
         oRef.Enabled:= lEnable
      endif
      oRef = oRef.before
   until oRef.Name == oParent.first.name
return
// EoF: EnableControls()

function Repaint
   /*
      This is a routine by Jim Sare with minor
      modifications by Ken Mayer. It can be used
      to ensure that if any processing has occurred
      that the form gets repainted (which sometimes
      doesn't happen) ... 

      Dec. 19, 1999: modified again by Romain Strieff,
                     basically a rewrite ...


      You can repaint a specific object on a form
      by passing the hWnd to the function, or
      you can call this from with the form's hWnd,
      or if you add this code into a form's 
      CLASS definition, just calling the method
      will use the form's hWnd by default.

      Ex: 
         class MyForm of form()
            // insert repaint here, before ENDCLASS
         endclass

      To use:
         for an object on the form:
            form.repaint( form.text1.hwnd )
         for the form:
            form.repaint()
   */
   if type("UpdateWindow") # "FP"
      extern cLogical UpdateWindow( cHandle ) user32
   endif
  
   if argcount()=1
      UpdateWindow(argVector(1))
   else
      UpdateWindow(form.hwnd)
   endif
return
// eof: repaint()

/*
   Repaint2()
   A different version of Repaint (see above) -- if you pass
   to this version a parameter, it will work on the object,
   if you pass the number zero (0), it will repaint the desktop
   (and everything on it). This might resolve some of the
   painting issues in 2.61.3 ...

   Posted in the dBASE Newsgroups by Rich Muller, April, 2007.

   Parameter: uH -- May be an object (i.e., form, container ...);
                    May be an object's hWnd (i.e., form.hwnd ...);
                    May be the number zero (to repaint the desktop)
*/
function Repaint2
   parameter uH
   local hWnd, hdl
   if type( "uH" ) == "O" // object
      hdl = uH.hWnd
   else
      hdl = uH
   endif
   // if the value is zero:
   if hdl == 0
      // we're going to get the handle for the desktop window:
      // prototype the function:
      if type( "GetDesktopWindow" ) # "FP"
         extern CHANDLE GetDesktopWindow() User32
      endif
      // get the handle
      hWnd = GetDesktopWindow()
   else
      hWnd = hdl
   endif
   // prototype and repaint the darn window:
   if type( "UpdateWindow" ) # "FP"
      extern CLOGICAL UpdateWindow( CHANDLE ) User32
   endif
   UpdateWindow( hWnd )
return true
// eof: repaint2()

function restore
   /*
      Restores a form that has been minimized ...
      Posted by Romain Strieff in the VdB newsgroups, Sept. 17, 1999
      Add to a form by copying and pasting into form's code
       before the form's "ENDCLASS" statement ... call as:
       form.restore()
   */
   if type( "SendMessage" ) # "FP"
      extern CLONG SendMessage(CHANDLE, CUINT, CLONG, CLONG) USER32;
             from "SendMessageA"
   endif

   #ifndef WM_SYSCOMMAND
      #define WM_SYSCOMMAND 0x0112
   #endif
   #ifndef SC_RESTORE
      #define SC_RESTORE    0xF120
   #endif

   SendMessage( form.hwnd, WM_SYSCOMMAND, SC_RESTORE, 0)
return
// end of Restore()

/*
   IsKeyPressed()
   Provided by Jim Sare in the dBASE newsgroups, June 23, 2003

   The following code is designed to allow you to 
   determine if either the SHIFT or CTRL keys are
   pressed when a form is opening, as the form in dBASE
   cannot check for this ... there are two parts, obviously,
   the first is code that is inserted into the form's onOpen
   event, you can of course modify it as needed, except for
   the #DEFINE statements -- those should be left alone.
   The second part is code that needs to be added to your
   form's code before the 'endclass' statement:

*/
    PROCEDURE form_onOpen

       #define VK_SHIFT          0x10
       #define VK_CONTROL        0x11
       if form.isKeyPressed(VK_SHIFT)
          ? "Shift key is down."
       else
          ? "Shift key is up."
       endIf
       if form.isKeyPressed(VK_CONTROL)
          ? "Ctrl key is down."
       else
          ? "Ctrl key is up."
       endIf

    FUNCTION IsKeyPressed(nKey)
       if type("GetKeyState") # "FP"
          extern CSHORT GetKeyState(CINT) user32 from "GetKeyState"
       endIf
    RETURN bitSet(GetKeyState(nKey), 15)
return
// EOF: IsKeyPressed()

FUNCTION Freeze
   /*
      Freeze -- freezes the current window, so you can
                modify the display without the form
                flashing. This is rather handy, combined
                with UnFreeze (below), and perhaps
                repaint().

      Credit: This code was lifted from the code in Command.wfm,
               which is unattributed ...
               
      Modified -- April 2016 by Mervyn Bick.
                  Original code only worked when called by an event handler
                  of a form or container and ignored any parameter passed to it.
                  Modified to work on the passed hwnd reference even when called,
                  for example, by a pushbutton's onClick event handler. If no
                  window handle is passed then it behaves as in the past and will
                  freeze the window that called it.
   */
   if type( "LockWindowUpdate" ) # "FP"
      extern CLOGICAL LockWindowUpdate( CHANDLE ) User32
   endif
   
   ******* Code added by MB
   PARAMETERS hwnd

   if argCount() = 0
      hwnd = this.hwnd
   endif
   ********
   
   if hWnd > 0 // modified by MB
      LockWindowUpdate( hWnd ) // modified by MB
   endif
return
// eof: freeze()

FUNCTION UnFreeze
   /*
       UnFreeze -- Un-Freezes the current window, so you can
                 modify the display without the form
                 flashing. This is rather handy, combined
                 with Freeze (above), and perhaps
                 repaint().

       Credit: This code was lifted from the code in Command.wfm,
               which is unattributed ...
   */
   if type( "LockWindowUpdate" ) # "FP"
      extern CLOGICAL LockWindowUpdate( CHANDLE ) User32
   endif
   LockWindowUpdate( 0 )
return
// eof: unfreeze()

function ReleaseQueries
   /*
      -----------------------------------------------------------------
      ReleaseQueries -- Purpose of this function is to find 
      all queries and stamp them out as completely as possible...

      Created by Jan Hoelterling, based on an idea by 
      Ken Mayer, to release query objects explicitly.
      The reason for this is that when closing a form,
      often the query objects are still active. This can
      cause some serious issues with memory management
      as well as any code that requires exclusive access
      to the tables used by the queries.

      It is suggested that a call to this code be added
      to the form's overridden release method:
         function form_Release
            class::ReleaseQueries()
          return ( super::release() )

      If using subforms you might want to do this the same way,
      or even in the form's release:
         function form_Release
            class::ReleaseQueries( form.MySubform )
            class::ReleaseQueries() // form
         return ( super::release() )

      Parameters: oRef     = object reference (form or subform)
                  bVerbose = display output in Command Window?
                             default is false. 
                             Usage requires you pass the first
                             parameter, i.e.,
                               class::ReleaseQueries( form, true )

      WARNING:
         If you use this in the form's release method, it may
         lock up dBASE. If that happens, try putting it in the
         form's onClose event, instead. This may be version
         dependent -- in dBASE Plus 2.61 (initial release),
         Jan found that in the form's overridden release method
         that it locked up his machine. In 2.61.2, I (Ken Mayer)
         found that it didn't, but this was on a simple test
         case. 
      -----------------------------------------------------------------
   */
   
   if pCount() => 1
      oRef = argVector(1)
   else
      oRef = form
   endif
   if pCount() < 1
      bVerbose = false
   else
      bVerbose = argVector(2)
   endif

   PRIVATE cTemp, xTemp
   LOCAL aa, cMember
 
   try
     aa = Enumerate(oRef)        // Enumerate the passed object
     cMember = aa.FirstKey        // Get first member name
     do while not empty( cMember )
        cTemp = "oRef." + cMember
        try
           xTemp = &cTemp.
           if type("xTemp") = "O" and xtemp.baseclassname = "QUERY"
              if bVerbose
                 ? cMember            // Display the member name
                 ?? aa[cMember] at 30       // Display the member type
                 ?? "Class: "+xtemp.classname at 50
                 ?? "BC: "+xtemp.baseclassname at 70
              endif

              // the important part of the code:
              xTemp.active = false
              release object &cTemp.
           endif
        catch (exception e)
           if bVerbose
              ? program(),e.Message,e.lineno
           endif
        endtry
        cMember = aa.NextKey( cMember )   // Get next member name
     enddo
   catch(exception e)
      if bVerbose
         ? program(),e.Message,e.lineno
      endif
   endtry

return
// eof: ReleaseQueries()

// end of file: FormStuf.prg

