/*
Description:

The function setLdriverTable() is used for setting the language driver in the
table header of a level 4, 5 or 7 dbf table to a new value. If the table has one
or more index files only the production index file will have it's language
driver ID changed correpondingly.

When the language driver has been set tables having an index are reindexed. 

The function includes a parameter for selecting whether or not table values
will be transliterated in accordance with the new language driver. For example,
when changing the table language driver from DB437US0 to DBWINWE0 and the
parameter bTranslit is set to false, the string 'Andr Mller' will be displayed
as 'Andr Mller' in the changed table; when the paramter is set to true, the 
name will be displayed correctly.

Notice: In the process of transliteration the original table is copied to a temp
table and values are appended from the temp table back to the original table.
If the original table contained records marked for deletion, said records will
be appended back to the original table if SET DELETED is OFF and they will still
be marked for deletion. If SET DELETED is ON, records marked for deletion will
not be appended back to the original table.


Syntax:

setLdriverTable(<tablename expC>, <ldrivername expC>, <transliteration expL>)

tablename expC			: The table name including extension, "Fish.dbf"
ldrivername expC  	: The short name of the table language driver, "BLLT1DA0"
transliteration expL	: true, 	transliteration will take place
							  false, transliteration will not take place 

Examples of use:

	setLdriverTable("Fish.dbf", "BLLT1DA0", true)
	setLdriverTable("Level5.dbf", "DBWINWE0", true)

Dependencies:  modistru.cc, api_structs.cc, dbf7File.cc, all in the duflp

Programmer:

	Ivar B. Jessen 

History:

22-10-2006 -- Original version
17-07-2008 -- Added code to assure that the value in an autionc field is 
				  not altered by the act of appending from Trans.dbf.		
02-08-2008 -- Modified code to test only version 7 tables for autoinc field.
				  Cleaned up code and renamed some variables.
              Added a notice in the table header dealing with effects of setting
              deleted on/off.

*/

function setLdriverTable

parameters cTable, cLdriver, bTranslit


// Create a copy of the original table if transliteration of data is required
	close tables

	if bTranslit = true
  		if file("Trans.dbf")
			drop table Trans
		endif
		copy table ( cTable ) to "Trans.dbf"
   	use
	endif

// Get Table version
	vers = getVersion()

// Get the load driver ID corresponding to parameter cLdriver
	ldID = getLdriverID(cLdriver)

	do case
   	case vers = 3					// dBASE III table not valid here
   		return  msgbox("dBASE III Table")
		case vers = 4 or vers = 5	// Level 4 or 5 table. Write ID in table header 
			writeLdriverID(ldID)		// and in mdx if present
	   case vers =	7					// Level7 table. Write new ldriverName in table
   	   writeLdriverName(ldID)	// header and in mdx if present
	   otherwise
      	return null
   endcase


	if bTranslit 
		// Zap original table for later appending values from saved table trans.dbf
      _app.databases[ 1 ].emptyTable( cTable )
      do case
      	case vers == 3 or vers == 4
         	appendFromTrans()
			case vers == 7
         	// Look for an autoinc field and if found, get its field name.
	     		fName = findAutoinc()
				do case
      	   	case empty( fName )				// No autoinc field 
 	      	   	appendFromTrans()
            	case not empty( fName )			// Has autoinc field
	   	   		// Get next AutoInc value
						set procedure to :duflp:dbf7File.cc additive
   	   	   	db7 = new dbf7File( cTable )
			         fNextAutoNumber = db7.getNextAutoIncValue()
	   			 	// Set the type of the field to "LONG".
						setFieldType(fName, "LONG")
						appendFromTrans()
					  	// Set the type of the autoinc field back to "AUTOINC"
						setFieldType(fName, "AUTOINC")
	   	   	  	db7.setNextAutoIncValue( fNextAutoNumber )
   	   	   	close procedure :duflp:dbf7File.cc
				endcase
         endcase
     	endif


function getversion
**********************
	t = new tabledef()
	t.tablename = ( cTable )
	t.load()
   vers = t.version
   release object t
   release t

return vers


function writeLdriverName
*****************************
// Blank out existing language driver name in table header
	f0 = new file()
	f0.open(cTable, "RW")
	f0.seek(0x020, 0)
	cNull = replicate(chr(0), 32)	
	f0.write(cNull)

// Write new language driver name in table header
	f0.seek(0x020, 0)
	f0.write(cLdriver)
	f0.close()
   release object f0
   release f0

   cMdx = substr(cTable, 1, rat(".", cTable)) + "mdx"
   if not file(cMdx)
   	return
   endif

// Set LdriverID in mdx file
   setLdriverIDinMdx(cMdx)
   
return


function writeLdriverID
**************************
// Write new language driver ID in table header
	f2 = new file()
	f2.open(cTable, "RW")
	f2.seek(0x1D, 0)
	f2.write(chr(htoi(ldID)))
	f2.close()
   release object f2
   release f2

   cMdx = substr(cTable, 1, rat(".", cTable)) + "mdx"
   if not file(cMdx)
	  	return
   endif

// Set LdriverID in mdx file
   setLdriverIDinMdx(cMdx)

return


function setLdriverIDinMdx(cMdx)
************************************
	f3 = new file()
	f3.open(cMdx, "RW")

// Get number of tags
	f3.seek(0x1C, 0)
	nTags = asc(f3.read(1))

// Get Block size adder
	f3.seek(0x16, 0)
	b1l = asc(f3.read(1))
	b1h = asc(f3.read(1))
	blockSizeAdder = b1h*256 + b1l
	blockOffset = blockSizeAdder * 2

	f3.seek(0x001F, 0)
	f3.write(chr(htoi(ldID)))
	f3.seek(0x0C0B, 0)
	f3.write(chr(htoi(ldID)))

	for i = 1 to nTags - 1
   	f3.seek(blockOffset - 1, 1)
   	f3.write(chr(htoi(ldID)))
	next

	f3.close()
   release object f3
   release f3

return  


function getLdriverID(cLd)
******************************
	cHex = ""
	for i = 1 to 8
		cHex += itoh(asc(substr(cLd, i, 1)))
	next

   cStr = ""
	cStr += "046A0A14000000007500050000000000"
	cStr += "00000000000000000000000000000000"
	cStr +=  cHex + "0000000000000000"
	cStr += "00000000000000000000000000000000"
   cStr += "00000000494400000000000000000000"
	cStr += "00000000000000000000000000000000"
	cStr += "00000000490400000000000000000000"
	cStr += "000000000D1A"

   f1 = new file() 
   f1.create("Dummy.dbf")
   f1.open("Dummy.dbf","RW")

   for n=1 to len(cStr) step 2
		f1.write(chr(htoi(substr(cStr,n,2))))
   next

   f1.close()
   release object f1
   release f1

	use dummy excl
	index on ID tag ID
	use

   if not file("Dummy.mdx")
   	// mdx file not created, return error null
   	return null
   endif

// Read language driver ID from Dummy.mdx 
	f4 = new file()
	f4.open("Dummy.mdx", "R")
	f4.seek(0x01F, 0)
	ldID = itoh(asc(f4.read(1)), 2)
	f4.close()
   release object f4
   release f4
	drop table Dummy

return ldID


function findAutoinc
**********************
local autoincName
autoincName = ""
ot = new TableDef()
ot.tableName = ( cTable )
ot.load()
for i = 1 to ot.fields.size
	if ot.fields[i].type = "AUTOINC"
    	autoincName = ot.fields[i].fieldName
      exit
   endif
next i 
release object ot
release ot

return autoincName


function setFieldType(cfName, cType)
*****************************************
use
set procedure to :duflp:api_structs.cc additive
set procedure to :duflp:modistru.cc additive
mstru = new modistruclass()
mstru.cTableName = ( cTable )
mstru.cFieldName = ( cfName )
mstru.NewFieldType = ( cType )
mstru.restructure() 
mstru.release()
close procedure :duflp:modistru.cc
close procedure :duflp:api_structs.cc

return


function appendFromTrans
************************

use ( cTable ) excl
append from ( "Trans.dbf" ) reindex
use
drop table Trans 

return
